@@ -1,13 +1,15 @@
 from __future__ import annotations
-from typing import List, Callable
+
 import base64
 import json
 import time
+from typing import List, Callable, Union
 
 import requests
 
 from .constants import API_BASE, USER_AGENT
-from .errors import AuthenticationException
+from .errors import AuthenticationException, MissingPasswordException, MissingEmailException, NotFoundException, \
+    MissingOTPException, IncorrectOTPException
 
 
 def jwt_expired(token: str) -> bool:
@@ -35,6 +37,8 @@ class HTBClient:
 
             from hackthebox import HTBClient
             client = HTBClient(email="user@example.com", password="S3cr3tP455w0rd!")
+    Attributes:
+        challenge_cooldown: Time when next download is allowed
 
     """
     # noinspection PyUnresolvedReferences
@@ -42,6 +46,7 @@ class HTBClient:
     _access_token: str = None
     _refresh_token: str = None
     _api_base: str = None
+    challenge_cooldown: int = 0
 
     def _refresh_access_token(self):
         """
@@ -50,16 +55,15 @@ class HTBClient:
         when the current one expires
 
         """
-        # TODO: Find a way to test this
         headers = {"User-Agent": USER_AGENT}
-        r = requests.post(API_BASE + "login/refresh", json={
+        r = requests.post(self._api_base + "login/refresh", json={
                 "refresh_token": self._refresh_token
             }, headers=headers)
         data = r.json()['message']
         self._access_token = data['access_token']
         self._refresh_token = data['refresh_token']
 
-    def do_request(self, endpoint, json_data=None, data=None, authorized=True) -> dict:
+    def do_request(self, endpoint, json_data=None, data=None, authorized=True, download=False, post=False) -> Union[dict, bytes]:
         """
 
         Args:
@@ -67,12 +71,14 @@ class HTBClient:
             json_data: Data to be sent in JSON format
             data: Data to be sent in application/x-www-form-urlencoded format
             authorized: If the request requires an Authorization header
+            download: If we are downloading raw data
+            post: Force POST request
         Returns:
-            The JSON response from the API
+            The JSON response from the API or the raw data (if `download` is set)
 
         """
         headers = {"User-Agent": USER_AGENT}
-        if authorized and self._api_base == API_BASE:
+        if authorized:
             # Don't use authorization if the API base URL isn't the real one -
             # i.e. we're running a test
             if jwt_expired(self._access_token):
@@ -80,33 +86,65 @@ class HTBClient:
             headers['Authorization'] = "Bearer " + self._access_token
         while True:
             if not json_data and not data:
-                r = requests.get(self._api_base + endpoint, headers=headers)
-            else:
-                r = requests.post(self._api_base + endpoint, json=json_data, data=data, headers=headers)
-            # Not sure on the exact ratelimit - loop until we don't get 429
-            if r.status_code == 429:
-                time.sleep(1)
+                if post:
+                    r = requests.post(self._api_base + endpoint, headers=headers, stream=download)
+                else:
+                    r = requests.get(self._api_base + endpoint, headers=headers, stream=download)
             else:
+                r = requests.post(
+                    self._api_base + endpoint, json=json_data, data=data, headers=headers, stream=download
+                )
+            if r.status_code != 429:
                 break
-        return r.json()
+            # Not sure on the exact ratelimit - loop until we don't get 429
+            else:
+                time.sleep(1)
+        if r.status_code == 404:
+            raise NotFoundException
+        if download:
+            return r.content
+        else:
+            return r.json()
 
-    def __init__(self, email: str = None, password: str = None, api_base: str = API_BASE):
+    def __init__(self, email: str = None, password: str = None, otp: str|int = None, api_base: str = API_BASE):
         self._api_base = api_base
         if not password and not email:
             print("Must give an authentication method")
             raise AuthenticationException
         elif password and not email:
-            print("Missing email")
-            raise AuthenticationException
+            raise MissingEmailException
         elif email and not password:
-            print("Missing password")
-            raise AuthenticationException
+            raise MissingPasswordException
         else:
             data = self.do_request("login", json_data={
                 "email": email, "password": password
             }, authorized=False)
             self._access_token = data['message']['access_token']
-            self._refresh_token = data['message']['access_token']
+            self._refresh_token = data['message']['refresh_token']
+            if data['message']['is2FAEnabled'] is True:
+                if otp is None:
+                    raise MissingOTPException
+                if type(otp) == int:
+                    # Optimistically try and create a string
+                    otp = f"{otp:06d}"
+                resp = self.do_request("2fa/login", json_data={
+                    "one_time_password": otp
+                })
+                if "correct" not in resp['message']:
+                    raise IncorrectOTPException
+
+    # noinspection PyUnresolvedReferences
+    def search(self, search_term: str) -> "Search":
+        """
+
+        Args:
+            search_term: The search term to pass to the API
+
+        Returns: A Search object with lists of the items retrieved from the API
+
+        """
+        from .search import Search
+        return Search(search_term, self)
 
     # noinspection PyUnresolvedReferences
     def get_machine(self, machine_id: int) -> "Machine":
@@ -184,7 +222,7 @@ class HTBClient:
         Args:
             endgame_id: The ID of the Endgame to fetch
 
-        Returns: A list of Endgame
+        Returns: An Endgame
 
         """
         from .endgame import Endgame
@@ -208,6 +246,39 @@ class HTBClient:
             endgames.append(Endgame(endgame, self, summary=True))
         return endgames
 
+    # noinspection PyUnresolvedReferences
+    def get_fortress(self, fortress_id: int) -> "Fortress":
+        """Requests an Fortress from the API
+
+        Args:
+            fortress_id: The ID of the Fortress to fetch
+
+        Returns: A Fortresse
+
+        """
+        from .fortress import Fortress
+        data = self.do_request(f"fortress/{fortress_id}")["data"]
+        return Fortress(data, self)
+
+    # noinspection PyUnresolvedReferences
+    def get_fortresses(self, limit: int = None) -> List["Fortress"]:
+        """Requests a list of Fortresses from the API
+
+        Args:
+            limit: The maximum number of Fortresses to fetch
+
+        Returns: A list of Fortresses
+
+        """
+        from .fortress import Fortress
+        data = self.do_request(f"fortresses")["data"]
+        fortresses = []
+        # For some  reason, the fortress list is in the format {"1": <fortress1>, "2": <fortress2>}
+        # instead of [<fortress1>, <fortress2>], meaning we have to sort it ourselves
+        for fortress_id, fortress in sorted(data.items())[:limit]:
+            fortresses.append(Fortress(fortress, self, summary=True))
+        return fortresses
+
     # noinspection PyUnresolvedReferences
     def get_user(self, user_id: int) -> "User":
         """
@@ -238,11 +309,11 @@ class HTBClient:
 
     # noinspection PyUnresolvedReferences
     def get_hof(self, vip: bool = False) -> "Leaderboard":
+        """
+        Returns: A Leaderboard of the top 100 Users
+        """
         from .leaderboard import Leaderboard
         from .user import User
-        """
-        Returns: A Leaderboard of Users
-        """
         endpoint = "rankings/users"
         if vip:
             endpoint += "?vip=1"
@@ -251,10 +322,10 @@ class HTBClient:
 
     # noinspection PyUnresolvedReferences
     def get_hof_countries(self) -> "Leaderboard":
+        """
+        Returns: A Leaderboard of the top 100 Countries
+        """
         from .leaderboard import Leaderboard, Country
-        """
-        Returns: A Leaderboard of Countries
-        """
         data = self.do_request("rankings/countries")['data']
         return Leaderboard(data, self, Country)
 
@@ -280,6 +351,41 @@ class HTBClient:
         data = self.do_request("rankings/universities")['data']
         return Leaderboard(data, self, University)
 
+    # noinspection PyUnresolvedReferences
+    def get_current_vpn_server(self, release_arena=False) -> "VPNServer":
+        """
+        Returns: The currently assigned `VPNServer`
+
+        Args:
+            release_arena: Get the current release arena VPN server
+        """
+        from .vpn import VPNServer
+        if release_arena:
+            data = self.do_request("connections/servers?product=release_arena")["data"]["assigned"]
+        else:
+            data = self.do_request("connections/servers?product=labs")["data"]["assigned"]
+        return VPNServer(data, self)
+
+    # noinspection PyUnresolvedReferences
+    def get_all_vpn_servers(self, release_arena=False) -> "List[VPNServer]":
+        """
+        Returns: A list of `VPNServer`s
+
+        Args:
+            release_arena: Use the release arena VPN servers
+        """
+        from .vpn import VPNServer
+        if release_arena:
+            data = self.do_request("connections/servers?product=release_arena")["data"]["options"]
+        else:
+            data = self.do_request("connections/servers?product=labs")["data"]["options"]
+        servers = []
+        for location in data.keys():  # 'EU'
+            for location_role in data[location].keys():  # 'EU - Free'
+                for server in data[location][location_role]["servers"].values():
+                    servers.append(VPNServer(server, self))
+        return servers
+
     # noinspection PyUnresolvedReferences
     @property
     def user(self) -> "User":
@@ -326,3 +432,6 @@ class HTBObject:
             return getattr(new_obj, item)
         else:
             raise AttributeError
+
+    def __eq__(self, other):
+        return self.id == other.id and type(self) == type(other)
